/*
 *   Xfrisk - The classic board game for X
 *   Copyright (C) 1993-1999 Elan Feingold (elan@aetherworks.com)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   $Id: callbacks.c,v 1.33 2003/07/14 10:37:38 sketch Exp $
 *
 *   Used by client/AI
 *   Notes:
 *     should not use X, so ai can be compiled on X-less box
 *      is included by riskgame.c, maybe fix there?
 *      or this be the X part, and move out other messages?
 */

static const char
    rcsid[] = "@(#)$Id: callbacks.c,v 1.33 2003/07/14 10:37:38 sketch Exp $";

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <X11/X.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/Toggle.h>

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

#include "language.h"
#include "network.h"
#include "gui-vars.h"
#include "gui-func.h"
#include "utils.h"
#include "network.h"
#include "callbacks.h"
#include "riskgame.h"
#include "client.h"
#include "dice.h"
#include "cards.h"
#include "game.h"
#include "colormap.h"
#include "help.h"
#include "colorEdit.h"
#include "registerPlayers.h"
#include "debug.h"
#include "viewStats.h"


/* Game globals -- move elsewhere? -- make part of riskgame object state! */
static Flag     fPlayingRemotely    = FALSE;
static Flag     fForceExchange      = FALSE;

static Int32    piMsgDstPlayerID[ MAX_PLAYERS + 1 ] = { 0 };
static Int32    iCountryToFortify   = -1;
static Flag     fWaitMission        = FALSE;
static Flag     fHaveSeenFirstPlayer;

Int32           iIndexMD;
Int32           iActionState        = ACTION_PLACE;
Int32           iReply;
CString         pstrMsgDstCString[ MAX_PLAYERS + 1 ];
Int32           iFirstPlayer;
Int32           iState;
Int32           iCurrentPlayer;
Flag            fGameStarted        = FALSE;

/**
 * X toolkit callback for the client-server communication socket
 */
void CBK_XIncomingMessage( XtPointer pClientData, int *iSource, XtInputId *id )
    {
    Msgs      *pMess;

    UNUSED( pClientData );
    UNUSED( id );

    RISK_ReceiveMessage( *iSource, &pMess );
    if ( pMess != NULL )
        {
        CBK_IncomingMessage( pMess );
        NET_DeleteMessage( pMess );
        }
    }


/**
 * Incoming message handler
 */
void CBK_IncomingMessage( const Msgs *pMess )
    {
    char    buf[ 256 ];
    Msgs   *msgCopy;
    Msgs    msg;
    Int32   iWinner;
    Int32   iPlayer;
#ifdef ASSERTIONS
    Int32   i;
#endif

    assert( pMess );
    switch ( pMess->msgId )
        {
        case MSG_NOMESSAGE:
            /* NoOp */
            break;

        case MSG_NETMESSAGE:
            assert(pMess->msgNetMessage.strMessage );
            UTIL_DisplayComment( pMess->msgNetMessage.strMessage );
            break;

        case MSG_PLACENOTIFY:
            /* Copy the message, since we are using it for a while */
            msgCopy = ( Msgs * ) MEM_Alloc( sizeof( Msgs ) );
            memcpy( ( void * ) msgCopy, ( void * ) pMess, sizeof( Msgs ) );

            /* Do the notification, and then add a timeout for erasing it */
            UTIL_PlaceNotification( ( XtPointer ) msgCopy, NULL );
            XtAppAddTimeOut( appContext,
                             NOTIFY_TIME,
                             UTIL_PlaceNotification,
                             msgCopy );
            break;

        case MSG_MOVENOTIFY:
            /* Copy the message, since we are using it for a while */
            msgCopy =  ( Msgs * ) MEM_Alloc( sizeof( Msgs ) );
            memcpy( ( void * ) msgCopy, ( void * ) pMess, sizeof( MsgMoveNotify ) );

            /* Do the notification, and then add a timeout for erasing it */
            UTIL_MoveNotification( ( XtPointer ) msgCopy, NULL );
            XtAppAddTimeOut( appContext,
                             NOTIFY_TIME,
                             UTIL_MoveNotification,
                             msgCopy );
            break;

        case MSG_ATTACKNOTIFY:
            /* Copy the message, since we are using it for a while */
            msgCopy =  ( Msgs * ) MEM_Alloc( sizeof( Msgs ) );
            memcpy( ( void * ) msgCopy, ( void * ) pMess, sizeof( Msgs ) );

            /* Do the notification, and then add a timeout for erasing it */
            UTIL_AttackNotification( ( XtPointer )msgCopy, NULL );
            XtAppAddTimeOut( appContext,
                             NOTIFY_TIME,
                             UTIL_AttackNotification,
                             msgCopy );
            break;

        case MSG_DICEROLL:
            /* Set the dice colors to be the correct ones */
            COLOR_CopyColor( COLOR_PlayerToColor( iCurrentPlayer ),
                             COLOR_DieToColor( 0 ) );
            COLOR_CopyColor( COLOR_PlayerToColor( pMess->msgDiceRoll.iDefendingPlayer ),
                             COLOR_DieToColor( 1 ) );

            /* Erase what was there, and then display the dice */
            DICE_Hide( );
            DICE_DrawDice( pMess->msgDiceRoll.pAttackDice,
                           pMess->msgDiceRoll.pDefendDice );
            break;

        case MSG_NETPOPUP:
            ( void )UTIL_PopupDialog( "Message from Server",
                                      pMess->msgNetPopup.strMessage,
                                      1,
                                      "Ok",
                                      NULL,
                                      NULL );
            break;

        case MSG_POPUPREGISTERBOX:
            /* Clear any text... */

            UTIL_DisplayError( "" );
            UTIL_DisplayComment("");

            REG_PopupDialog( );
            break;

        case MSG_FORCEEXCHANGECARDS:
            if ( iState != STATE_REGISTER )
                {
                if ( pMess->msgForceExchangeCards.iPlayer >= 0 )
                    {
                    UTIL_DisplayComment( "You have too many cards"
                                         " and must exchange a set." );
                    fForceExchange = fCanExchange = TRUE;
                    CBK_ShowCards( ( Widget ) NULL, ( XtPointer ) NULL, ( XtPointer ) NULL );

                    iState = STATE_PLACE;
                    UTIL_ServerEnterState( iState );
                    }
                }
            break;

        case MSG_REPLYPACKET:
            iReply = pMess->msgReplyPacket.iReply;
            break;

        case MSG_ENDOFMISSION:
            iWinner = pMess->msgEndOfMission.iWinner;

            assert( iWinner >= 0 && iWinner < MAX_PLAYERS );

            /* Print out a final message */
            GAME_MissionAccomplied( iWinner,
                                    pMess->msgEndOfMission.iTyp,
                                    pMess->msgEndOfMission.iNum1,
                                    pMess->msgEndOfMission.iNum2 );
            break;

        case MSG_VICTORY:
            iWinner = pMess->msgVictory.iWinner;

            assert( iWinner >= 0 && iWinner < MAX_PLAYERS );

            /* Print out a final message */
            GAME_Victory( iWinner );
            break;

        case MSG_ENDOFGAME:
            /* Set the cursor so that the client doesn't have
             * the icon of the hourglass by any chance, because
             * it doesn't really make sense.
             */

            UTIL_SetCursorShape( CURSOR_PLAY );

            /* Print out a final message */
            fHaveSeenFirstPlayer = FALSE;
            GAME_GameOverMan( );
            break;

        case MSG_MESSAGEPACKET:
            UTIL_DisplayMessage( pMess->msgMessagePacket.iFrom,
                                 pMess->msgMessagePacket.iTo,
                                 pMess->msgMessagePacket.strMessage );
            break;

        case MSG_TURNNOTIFY:
            iCurrentPlayer = pMess->msgTurnNotify.iPlayer;
            if ( !fHaveSeenFirstPlayer )
                {
                fHaveSeenFirstPlayer = TRUE;
                iFirstPlayer = iCurrentPlayer;
                }

            /* Set the color of the player color-coded indicator */
            GUI_SetColorOfCurrentPlayer( COLOR_PlayerToColor( iCurrentPlayer ) );

            if ( pMess->msgTurnNotify.iClient == CLNT_GetThisClientID( ) )
                {
                fPlayingRemotely = FALSE;

                /* Set the cursor to indicate play */
                UTIL_SetCursorShape( CURSOR_PLAY );

                /* See if everyone has finished fortifying their territories.
                 * If they have not, then there is a serious problem.
                 */

                if ( !fGameStarted && ( iState == STATE_FORTIFY ) &&
                     ( RISK_GetNumArmiesOfPlayer( iCurrentPlayer ) == 0 ) )
                    {
                    fGameStarted = TRUE;
                    iCountryToFortify = -1;

#ifdef ASSERTIONS
                    /* Sanity check */
                    for ( i = 0;
                          i != RISK_GetNumLivePlayers( );
                          i++ )
                        {
                        Int32 iPlayer = RISK_GetNthLivePlayer( i );
                        Int32 iFirstAttacker;
                        Int32 j;

                        assert( RISK_GetNumArmiesOfPlayer( iPlayer ) >= 0 );

                        assert( RISK_GetNumArmiesOfPlayer( iPlayer ) == 0 );

                        /* I not entirely sure this correct. But it's at least a
                         * little better than before. People who have already had
                         * their first turn are allowed to have cards. --Pac */
                        for (  j = 0; j != RISK_GetNumLivePlayers( ); ++j )
                            {
                            if ( RISK_GetNthLivePlayer( j )==iFirstPlayer )
                                {
                                break;
                                }
                            }
                        assert( j != RISK_GetNumLivePlayers( ) );

                        /* negative % positive = negative, unfortunately */
                        iFirstAttacker = j - NUM_COUNTRIES%RISK_GetNumLivePlayers( );
                        while ( iFirstAttacker < 0 )
                            {
                            iFirstAttacker += RISK_GetNumLivePlayers( );
                            }
                        iFirstAttacker = RISK_GetNthLivePlayer( iFirstAttacker );
                        assert( iPlayer < iCurrentPlayer
                                || iPlayer >= iFirstAttacker
                                || RISK_GetNumCardsOfPlayer( iPlayer ) == 0 );
                        }
#endif
                    }

                if ( fGameStarted )
                    {
                    /* Force the player to exchange if he or she possesses
                     * 5 cards or more.
                     */

                    if ( RISK_GetNumCardsOfPlayer( iCurrentPlayer ) >= 5 )
                        {
                        UTIL_DisplayComment( "You have more than five cards"
                                             " and must exchange a set." );
                        fForceExchange = TRUE;
                        CBK_ShowCards( ( Widget ) NULL,
                                       ( XtPointer ) NULL,
                                       ( XtPointer ) NULL );
                        }

                    /* Go into placing state, give the player as many armies as
                     * he or she deserves by dividing the total number of
                     * countries owned divided by 3, for a minimum of three
                     * armies.
                     */

                    iState = STATE_PLACE;
                    UTIL_ServerEnterState( iState );

                    /* All of the armies should have been used up */
                    assert( RISK_GetNumArmiesOfPlayer( iCurrentPlayer ) == 0 );

                    GAME_SetTurnArmiesOfPlayer( iCurrentPlayer );
                    }
                else if ( iState == STATE_FORTIFY )
                    {
                    UTIL_ServerEnterState( iState );
                    if ( iCountryToFortify != -1 )
                        {
                        GAME_PlaceClick( iCountryToFortify, PLACE_ONE );
                        iCountryToFortify = -1;
                        UTIL_DisplayError( "" );
                        }
                    }
                UTIL_DisplayActionCString( iState, iCurrentPlayer );
                }
            else
                {
                fPlayingRemotely = TRUE;

                /* Set the cursor to indicate waiting */
                UTIL_SetCursorShape( CURSOR_WAIT );

                snprintf( buf, sizeof( buf ), "%s is playing remotely...",
                          RISK_GetNameOfPlayer( iCurrentPlayer ) );
                UTIL_DisplayComment( buf );
                }
            break;

        case MSG_EXIT:
            /* Popup telling what happened */
            ( void )UTIL_PopupDialog( "Exit Notification",
                                      "Terminating game!  Goodbye...",
                                      1, "Ok", NULL, NULL );
            close( CLIENTS_GetCommLinkOfClient( CLNT_GetThisClientID( ) ) );
            CLIENTS_SetCommLink( CLNT_GetThisClientID( ), -1 );
            UTIL_ExitProgram( 0 );
            break;

        case MSG_MISSION:
            if ( !fWaitMission )
                {
                if ( UTIL_PlayerIsLocal( iCurrentPlayer ) )
                    {
                    iPlayer = iCurrentPlayer;
                    }
                else if ( UTIL_NumPlayersAtClient( CLNT_GetThisClientID( ) ) == 1 )
                    {
                    iPlayer = RISK_GetNthPlayerAtClient( CLNT_GetThisClientID( ), 0 );
                    }
                else
                    {
                    iPlayer = -1;
                    }
                if ( iPlayer != -1 )
                    {
                    GAME_ShowMission( iPlayer );
                    ( void )UTIL_PopupDialog( "Mission Notification",
                                              "The mission is visible in the "
                                              "message zone",
                                              1,
                                              "Ok",
                                              NULL,
                                              NULL );
                    }
                else
                    {
                    /* Popup telling what happened */
                    ( void )UTIL_PopupDialog( "Mission Notification",
                                              "To view your mission, click on\n"
                                              " mission's button",
                                              1,
                                              "Ok",
                                              NULL,
                                              NULL );
                    }
                }
            break;

        default:
            snprintf( buf,
                      sizeof( buf ), "CALLBACKS: Unhandled message( %d )",
                      pMess->msgId );
            ( void ) UTIL_PopupDialog( "Fatal Error",
                                       buf,
                                       1,
                                       "Ok",
                                       NULL,
                                       NULL );
            msg.msgId = MSG_DEREGISTERCLIENT;
            RISK_SendMessage( CLIENTS_GetCommLinkOfClient( CLNT_GetThisClientID( ) ),
                              &msg );
            UTIL_ExitProgram( -1 );
        }
    }


/**
 * Map refresh callback
 */
void CBK_RefreshMap( Widget w, XtPointer pData, XtPointer pCalldata )
    {
    XExposeEvent *pExpose = ( XExposeEvent * )pCalldata;
    UNUSED( w );
    UNUSED( pData );

    XCopyArea( hDisplay,
               pixMapImage,
               hWindow,
               hGC,
               pExpose->x,
               pExpose->y,
               pExpose->width,
               pExpose->height,
               pExpose->x,
               pExpose->y );
    }


/**
 * Handler for mouse clicks on the map
 */
void CBK_MapClick( Widget w, XEvent *pEvent, String *str, Cardinal *card )
    {
    Int32  x, y, iCountry;
    char buf[ 256 ];
    UNUSED( w );
    UNUSED( str );
    UNUSED( card );

    if ( iState == STATE_REGISTER )
        {
        return;
        }

    /* Find out the coordinates and time of the mouse click */
    x = pEvent->xbutton.x;
    y = pEvent->xbutton.y;

    /* Find out what country was clicked on */
    iCountry = COLOR_ColorToCountry( XGetPixel( pMapImage, x, y ) );

    if ( fPlayingRemotely )
        {
        if ( ( UTIL_NumPlayersAtClient( CLNT_GetThisClientID( ) ) == 1 )
             && ( iState == STATE_FORTIFY )
             && ( iCountry < NUM_COUNTRIES ) )
            {
            if ( iCountryToFortify == -1 )
                {
                iCountryToFortify = iCountry;
                snprintf( buf,
                          sizeof( buf ),
                          "%s %s",
                          RISK_GetNameOfCountry( iCountryToFortify ),
                          TXT_BE_FORTIFIED );
                UTIL_DisplayError( buf );
                return;
                }
            }
        snprintf( buf,
                  sizeof( buf ),
                  "Click as you may, it's %s's turn...",
                  RISK_GetNameOfPlayer( iCurrentPlayer ) );
        UTIL_DisplayError( buf );
        return;
        }

    /* Erase previous errors */
    UTIL_DisplayError( "" );

    switch ( iState )
        {

        case STATE_FORTIFY:
        case STATE_PLACE:
            if ( pEvent->xbutton.button == Button1 )
                {
                GAME_PlaceClick( iCountry, PLACE_ONE );
                }
            else if ( pEvent->xbutton.button == Button2 )
                {
                GAME_PlaceClick( iCountry, PLACE_ALL );
                }
            else if ( pEvent->xbutton.button == Button3 )
                {
                GAME_PlaceClick( iCountry, PLACE_MULTIPLE );
                }
            break;

        case STATE_ATTACK:
            GAME_AttackClick( iCountry );
            break;

        case STATE_MOVE:
            GAME_MoveClick( iCountry );
            break;

        default:
            assert( FALSE );
        }
    }


/**
 * Quit button handler
 */
void CBK_Quit( Widget w, XtPointer pData, XtPointer call_data )
    {
    Int32 rep;
    Msgs  msg;

    UNUSED( w );
    UNUSED( pData );
    UNUSED( call_data );

    rep = UTIL_PopupDialog( "Quit",
                            "Quit Xfrisk or new game?",
                            3,
                            "Quit",
                            "New",
                            "Cancel" );
    switch ( rep )
        {
        case QUERY_YES:
            UTIL_ExitProgram( 0 );
        case QUERY_NO:
            iState = STATE_REGISTER;
            msg.msgId = MSG_ENDOFGAME;
            ( void ) RISK_SendMessage( CLIENTS_GetCommLinkOfClient( CLNT_GetThisClientID( ) ),
                                       &msg );
        }
    }


/**
 * Show cards button handler
 */
void CBK_ShowCards( Widget w, XtPointer pData, XtPointer call_data )
    {
    Int32 i;
    Int32 iPlayer;
    UNUSED( w );
    UNUSED( pData );
    UNUSED( call_data );

    if ( iState == STATE_REGISTER )
        {
        return;
        }

    UTIL_DisplayError( "" );

    /* If the player clicking is the current player, show cards, otherwise,
     * if there's more than one player at the client, issue an error, unless
     * there's only one, in which case show his or her cards.
     */

    if ( UTIL_PlayerIsLocal( iCurrentPlayer ) )
        {
        iPlayer = iCurrentPlayer;
        }
    else if ( UTIL_NumPlayersAtClient( CLNT_GetThisClientID( ) ) == 1 )
        {
        iPlayer = RISK_GetNthPlayerAtClient( CLNT_GetThisClientID( ), 0 );
        }
    else
        {
        iPlayer = -1;
        ( void )UTIL_PopupDialog( "Error",
                                  "Wait until your turn!",
                                  1,
                                  "Ok",
                                  "Cancel",
                                  NULL );
        }

    /* If it's valid, display cards */
    if ( iPlayer >= 0 )
        {
        XtPopup( wCardShell, XtGrabExclusive );

        for ( i = 0; i != RISK_GetNumCardsOfPlayer( iPlayer ); i++ )
            {
            CARDS_RenderCard( RISK_GetCardOfPlayer( iPlayer, i ), i );
            }
        }
    }


/**
 * Cards view cancel button handler
 */
void CBK_CancelCards( Widget w, XtPointer pData, XtPointer call_data )
    {
    Int32 i;
    UNUSED( w );
    UNUSED( pData );
    UNUSED( call_data );

    if ( fForceExchange == TRUE )
        ( void ) UTIL_PopupDialog( "Error",
                                   "You must exchange cards!",
                                   1,
                                   "Ok",
                                   NULL,
                                   NULL );
    else
        {
        UTIL_DisplayError( "" );
        XtPopdown( wCardShell );

        /* Unmap the cards and clear the selections */
        for ( i = 0; i != MAX_CARDS; i++ )
            {
            XtUnmanageChild( wCardToggle[ i ] );
            XtVaSetValues( wCardToggle[ i ],
                           XtNstate,
                           False,
                           NULL );
            }
        }
    }


/**
 * Show mission handler
 */
void CBK_ShowMission( Widget w, XtPointer pData, XtPointer call_data )
    {
    Int32 iPlayer;
    Msgs  msg;

    UNUSED( w );
    UNUSED( pData );
    UNUSED( call_data );

    if ( iState == STATE_REGISTER )
        {
        return;
        }

    UTIL_DisplayError( "" );

    /* If the player clicking is the current player, show cards, otherwise,
     * if there's more than one player at the client, issue an error, unless
     * there's only one, in which case show his or her cards.
     */

    if ( UTIL_PlayerIsLocal( iCurrentPlayer ) )
        {
        iPlayer = iCurrentPlayer;
        }
    else if ( UTIL_NumPlayersAtClient( CLNT_GetThisClientID( ) ) == 1 )
        {
        iPlayer = RISK_GetNthPlayerAtClient( CLNT_GetThisClientID( ), 0 );
        }
    else
        {
        iPlayer = -1;
        ( void ) UTIL_PopupDialog( "Error",
                                   "Wait until your turn!",
                                   1,
                                   "Ok",
                                   NULL,
                                   NULL );
        }

    /* If it's valid, display mission */
    if ( iPlayer >= 0 )
        {
        if ( RISK_GetMissionTypeOfPlayer( iPlayer ) == NO_MISSION )
            {
            if ( UTIL_PopupDialog( "Mission",
                                   "Play with a mission?",
                                   2,
                                   "Ok",
                                   "Cancel",
                                   NULL ) == QUERY_NO )
                    return;
            fWaitMission = TRUE;
            msg.msgId = MSG_MISSION;
            ( void ) RISK_SendSyncMessage( CLIENTS_GetCommLinkOfClient( CLNT_GetThisClientID( ) ),
                                           &msg,
                                           MSG_MISSION,
                                           CBK_IncomingMessage );
            fWaitMission = FALSE;
            }
        GAME_ShowMission( iPlayer );
        }
    }


/**
 * Attack handler
 */
void CBK_Attack( Widget w, XtPointer pData, XtPointer call_data )
    {
    XawListReturnStruct *pItem;
    UNUSED( w );
    UNUSED( pData );
    UNUSED( call_data );

    if ( iState == STATE_REGISTER )
        {
        return;
        }

    /*
     * Don't do anything if current player is not local and there is not
     * one player at this client.
     */
    if ( UTIL_PlayerIsLocal( iCurrentPlayer ) == FALSE )
        {
        return;
        }

    /* I know this is silly, but hey... ( We should get the data from pData? ) */
    pItem = XawListShowCurrent( wAttackList );
    RISK_SetDiceModeOfPlayer( iCurrentPlayer, pItem->list_index );

    /* Free up memory */
    XtFree( ( void * ) pItem );
    }


/**
 * Action handler
 */
void CBK_Action( Widget w, XtPointer pData, XtPointer call_data )
    {
    XawListReturnStruct *pItem;
    UNUSED( w );
    UNUSED( pData );
    UNUSED( call_data );

    if ( iState == STATE_REGISTER )
        {
        return;
        }

    /* Don't do anything if current player is not local */
    if ( !UTIL_PlayerIsLocal( iCurrentPlayer ) )
        {
        return;
        }

    /* I know this is silly, but hey... ( We should get the data from pData? ) */
    pItem = XawListShowCurrent( wActionList );

    switch ( pItem->list_index )
        {
        case ACTION_PLACE:
            if ( iState == STATE_PLACE || iState == STATE_FORTIFY )
                {
                UTIL_DisplayError( "" );
                iActionState = pItem->list_index;
                }
            else /* Invalid */
                {
                UTIL_DisplayError( "You've already placed your armies." );
                XawListHighlight( wActionList, ACTION_ATTACK );
                }

            break;

        case ACTION_ATTACK:
        case ACTION_DOORDIE:
            if ( iState == STATE_ATTACK )
                {
                UTIL_DisplayError( "" );
                iActionState = pItem->list_index;

                /* Keep track of the type of attack selected */
                RISK_SetAttackModeOfPlayer( iCurrentPlayer, pItem->list_index );
                }
            else if ( iState == STATE_PLACE || iState == STATE_FORTIFY )
                {
                UTIL_DisplayError( "You must finish placing your armies." );
                XawListHighlight( wActionList, ACTION_PLACE );
                }
            else if ( iState == STATE_MOVE ) /* Must not have moved yet */
                {
                /* Keep track of the type of attack selected */
                RISK_SetAttackModeOfPlayer( iCurrentPlayer, pItem->list_index );

                iState = STATE_ATTACK;
                UTIL_DisplayError( "" );
                UTIL_ServerEnterState( iState );
                UTIL_DisplayActionCString( iState, iCurrentPlayer );
                }

            break;

        case ACTION_MOVE:
            if ( iState == STATE_MOVE )
                {
                UTIL_DisplayError( "" );
                iActionState = pItem->list_index;
                }
            else if ( iState == STATE_FORTIFY || iState == STATE_PLACE )
                {
                UTIL_DisplayError( "You must finish placing your armies." );
                XawListHighlight( wActionList, ACTION_PLACE );
                }
            else /* STATE_ATTACK */
                {
                iState = STATE_MOVE;
                UTIL_DisplayError( "" );
                UTIL_ServerEnterState( iState );
                UTIL_DisplayActionCString( iState, iCurrentPlayer );
                }
            break;

        default:
            assert( FALSE );
        }

    /* Free up memory */
    XtFree( ( void * ) pItem );
    }


/**
 * Message destination selection change handler
 */
void CBK_MsgDest( Widget w, XtPointer pData, XtPointer call_data )
    {
    XawListReturnStruct *pItem;
    Int32 iPlayer;
    UNUSED( w );
    UNUSED( pData );
    UNUSED( call_data );

    if ( UTIL_PlayerIsLocal( iCurrentPlayer ) )
        {
        iPlayer = iCurrentPlayer;
        }
    else if ( UTIL_NumPlayersAtClient( CLNT_GetThisClientID( ) ) == 1 )
        {
        iPlayer = RISK_GetNthPlayerAtClient( CLNT_GetThisClientID( ), 0 );
        }
    else
        {
        iPlayer = -1;
        }

    if ( iPlayer == -1 )
        {
        return;
        }

    /* I know this is silly, but hey... ( We should get the data from pData? ) */
    pItem = XawListShowCurrent( wMsgDestList );
    RISK_SetMsgDstModeOfPlayer( iPlayer, pItem->list_index );

    /* Free up memory */
    XtFree( ( void * ) pItem );
    }


/**
 * Message send handler
 */
void CBK_SendMessage( void )
    {
    CString                strMess;
    XawListReturnStruct   *pItem;
    Msgs                   mess;

    /* Get the message and destination from widgets */
    XtVaGetValues( wSendMsgText,
                   XtNstring,
                   &strMess,
                   NULL );
    pItem = XawListShowCurrent( wMsgDestList );

    /* Who's sending the message? */
    if ( UTIL_NumPlayersAtClient( CLNT_GetThisClientID( ) ) == 1 )
        {
        mess.msgMessagePacket.iFrom = RISK_GetNthPlayerAtClient( CLNT_GetThisClientID( ), 0 );
        }
    else if ( RISK_GetClientOfPlayer( iCurrentPlayer ) == CLNT_GetThisClientID( ) )
        {
        mess.msgMessagePacket.iFrom = iCurrentPlayer;
        }
    else
        {
        mess.msgMessagePacket.iFrom = FROM_UNKNOW;
        }

    /* Who do we send the message to? */
    if ( pItem->list_index == -1 || pItem->list_index == 0 )
        {
        mess.msgMessagePacket.iTo = DST_ALLPLAYERS;
        }
    else
        {
        mess.msgMessagePacket.iTo = piMsgDstPlayerID[ pItem->list_index ];
#if 0
        assert( mess.msgMessagePacket.iTo!=-1 );
#endif
        }

    mess.msgMessagePacket.strMessage = strMess;

    /* Display the message locally if necessary */
    if ( ( mess.msgMessagePacket.iTo != DST_ALLPLAYERS )
         && ( RISK_GetClientOfPlayer( mess.msgMessagePacket.iTo ) !=
              CLNT_GetThisClientID( ) ) )
        {
        UTIL_DisplayMessage( mess.msgMessagePacket.iFrom,
                             mess.msgMessagePacket.iTo,
                             mess.msgMessagePacket.strMessage );
        }
    /* Send the message */
    mess.msgId = MSG_MESSAGEPACKET;
    ( void ) RISK_SendMessage( CLIENTS_GetCommLinkOfClient( CLNT_GetThisClientID( ) ),
                               &mess );

    /* Erase the old message */
    XtVaSetValues( wSendMsgText, XtNstring, "", NULL );

    /* Free up memory */
    XtFree( ( void * ) pItem );
    }


/**
 * Cancel attack handler
 */
void CBK_CancelAttack( Widget w, XtPointer pData, XtPointer call_data )
    {
    UNUSED( w );
    UNUSED( pData );
    UNUSED( call_data );

    if ( iState == STATE_REGISTER )
        {
        return;
        }

    if ( fPlayingRemotely )
        {
        UTIL_DisplayError( "Nice try, but it isn't your turn..." );
        return;
        }

    UTIL_DisplayError( "" );

    if ( iState == STATE_ATTACK )
        {
        if ( GAME_AttackFrom( ) >=0 )
            {
            UTIL_DarkenCountry( GAME_AttackFrom( ) );
            }
        GAME_SetAttackSrc( -1 );
        }
    else if ( iState == STATE_MOVE )
        {
        if ( iMoveSrc >=0 )
            {
            UTIL_DarkenCountry( iMoveSrc );
            }
        iMoveSrc = -1;
        }

    UTIL_DisplayActionCString( iState, iCurrentPlayer );
    }


/**
 * Repeat attack handler
 */
void CBK_Repeat( Widget w, XtPointer pData, XtPointer call_data )
    {
    MsgNetMessage mess;
    char buf[ 256 ];
    UNUSED( w );
    UNUSED( pData );
    UNUSED( call_data );

    if ( iState == STATE_REGISTER )
        {
        return;
        }

    if ( fPlayingRemotely )
        {
        snprintf( buf,
                  sizeof( buf ),
                  "I'm afraid that's up to %s to decide...",
                  RISK_GetNameOfPlayer( iCurrentPlayer ) );
        UTIL_DisplayError( buf );
        return;
        }

    if ( iState != STATE_ATTACK )
        {
        UTIL_DisplayError( "You can't do any attacking at the moment." );
        }
    else if ( iActionState == ACTION_ATTACK || iActionState == ACTION_DOORDIE )
        {
        if ( iLastAttackSrc == -1 || iLastAttackDst == -1 )
            {
            UTIL_DisplayError( "You can't repeat an uncomplete attack." );
            return;
            }

        /* send a verbose message */
        snprintf( buf,
                  sizeof( buf ),
                  "%s is attacking from %s to %s",
                  RISK_GetNameOfPlayer( iCurrentPlayer ),
                  RISK_GetNameOfCountry( iLastAttackSrc ),
                  RISK_GetNameOfCountry( iLastAttackDst ) );
        mess.strMessage = buf;
        mess.msgId = MSG_NETMESSAGE;
        ( void ) RISK_SendMessage( CLIENTS_GetCommLinkOfClient( CLNT_GetThisClientID( ) ),
                                   ( const Msgs * ) &mess );

        /* Repeat the attack */
        GAME_Attack( iLastAttackSrc, iLastAttackDst );
        }
    }


/**
 * Help button handler
 */
void CBK_Help( Widget w, XtPointer pData, XtPointer call_data )
    {
    Int32 x, y;
    UNUSED( w );
    UNUSED( pData );
    UNUSED( call_data );

    /* Center popup */
    UTIL_CenterShell( wHelpShell,
                      wToplevel,
                      &x,
                      &y );

    XtVaSetValues( wHelpShell,
                   XtNallowShellResize,
                   False,
                   XtNx,
                   x,
                   XtNy,
                   y,
                   XtNborderWidth,
                   1,
                   XtNtitle,
                   "Xfrisk Help",
                   NULL );

    XtPopup( wHelpShell, XtGrabNone );
    }


/**
 * Help topic selection handler
 */
void CBK_HelpSelectTopic( Widget w, XtPointer pData, XtPointer call_data )
    {
    XawListReturnStruct *pItem;
    UNUSED( w );
    UNUSED( pData );
    UNUSED( call_data );

    /* I know this is sill, but hey... ( We should get the data from pData ) */
    pItem = XawListShowCurrent( wHelpTopicList );
    HELP_IndexPopupHelp( pItem->list_index );

    /* Free up memory */
    XtFree( ( void * )pItem );
    }


/**
 * Help OK button handler
 */
void CBK_HelpOk( Widget w, XtPointer pData, XtPointer call_data )
    {
    UNUSED( w );
    UNUSED( pData );
    UNUSED( call_data );
    XtPopdown( wHelpShell );
    }


/**
 * Card exchange handler
 */
void CBK_ExchangeCards( Widget w, XtPointer pData, XtPointer call_data )
    {
    Int32   i;
    Int32   iNumCardsSelected;
    Flag    iCardState;
    Int32   piCards[ 3 ];
    Int32   piCardValues[ 3 ];
    Int32   piCardTypes[ 3 ];
    UNUSED( w );
    UNUSED( pData );
    UNUSED( call_data );

    /* If it's not the current player, then don't let the exchange
     * take place.  This fixes a nasty bug reported by many.
     */

    if ( RISK_GetClientOfPlayer( iCurrentPlayer ) != CLNT_GetThisClientID( ) )
        {
        UTIL_DisplayError( TXT_ONLY_CURRENT );
        }

    /* Get the cards that have been pressed, check for validity */
    for ( i = iNumCardsSelected = 0; i != MAX_CARDS; i++ )
        {
        XtVaGetValues( wCardToggle[ i ],
                       XtNstate,
                       &iCardState,
                       NULL );
        if ( iCardState == True )
            {
            if ( iNumCardsSelected < 3 )
                {
                piCards[ iNumCardsSelected ] = i;
                }
            iNumCardsSelected++;
            }
        }

    /* Clear the selections of the cards */
    for ( i = 0; i != MAX_CARDS; i++ )
        {
        XtVaSetValues( wCardToggle[ i ],
                       XtNstate,
                       False,
                       NULL );
        }

    /* Valid number of cards selected */
    if ( fCanExchange == FALSE )
        {
        UTIL_DisplayError( "You can only exchange cards at the beginning of "
                           "your turn." );
        return;
        }
    else if ( iNumCardsSelected != 3 )
        {
        UTIL_DisplayError( "You must select three cards to exchange." );
        return;
        }

    /* Substitute the indices for the card values */
    for ( i = 0; i != 3; i++ )
        {
        piCardValues[ i ] = RISK_GetCardOfPlayer( iCurrentPlayer, piCards[ i ] );

        /* Set the type of the card */
        if ( piCardValues[ i ] < NUM_COUNTRIES )
            {
            piCardTypes[ i ] = piCardValues[ i ] % 3;
            }
        else
            {
            piCardTypes[ i ] = -1;  /* Joker */
            }
        }

    /* Now see if the cards form a valid triple */
    if ( ( piCardTypes[ 0 ] == piCardTypes[ 1 ] &&
           piCardTypes[ 1 ] == piCardTypes[ 2 ] &&
           piCardTypes[ 0 ] == piCardTypes[ 2 ] ) ||
         ( piCardTypes[ 0 ] != piCardTypes[ 1 ] &&
           piCardTypes[ 1 ] != piCardTypes[ 2 ] &&
           piCardTypes[ 0 ] != piCardTypes[ 2 ] ) ||
         ( piCardTypes[ 0 ] == -1 ||
           piCardTypes[ 1 ] == -1 ||
           piCardTypes[ 2 ] == -1 ) )
        {
        /* Actually perform the exchange */
        GAME_ExchangeCards( piCards );

        /* If this is the only exchange, then pop down the window */
        if ( RISK_GetNumCardsOfPlayer( iCurrentPlayer ) <= 5 )
            {
            /* We're through exchanging sets of cards */
            fForceExchange = FALSE;

            XtPopdown( wCardShell );

            /* Unmap the cards and clear the selections */
            for ( i = 0; i != MAX_CARDS; i++ )
                {
                XtUnmanageChild( wCardToggle[ i ] );
                }
            }
        else
            {
            /* Redisplay cards after erasing and unselecting the old ones */
            for ( i = 0; i != MAX_CARDS; i++ )
                {
                XtUnmanageChild( wCardToggle[ i ] );
                XtVaSetValues( wCardToggle[ i ],
                               XtNstate,
                               False,
                               NULL );
                }

            for ( i = 0; i != RISK_GetNumCardsOfPlayer( iCurrentPlayer ); i++ )
                {
                CARDS_RenderCard( RISK_GetCardOfPlayer( iCurrentPlayer, i ), i );
                }
            }

        /* Display the appropriate message for the state */
        UTIL_DisplayActionCString( iState, iCurrentPlayer );
        }
    else
        {
        UTIL_DisplayError( "You must pick three cards of the same type or one of "
                           "each type." );
        }
    }


/**
 * End turn handler
 */
void CBK_EndTurn( Widget w, XtPointer pData, XtPointer call_data )
    {
    char buf[ 256 ];
    UNUSED( w );
    UNUSED( pData );
    UNUSED( call_data );

    if ( iState == STATE_REGISTER )
        {
        return;
        }

    if ( fPlayingRemotely )
        {
        snprintf( buf,
                  sizeof( buf ),
                  "Come on, let %s finish playing...",
                  RISK_GetNameOfPlayer( iCurrentPlayer ) );
        UTIL_DisplayError( buf );
        return;
        }

    if ( iState == STATE_PLACE || iState == STATE_FORTIFY )
        {
        UTIL_DisplayError( "You must finish placing your armies." );
        }
    else
        {
        UTIL_DisplayError( "" );
        GAME_EndTurn( );
        }
    }


/**
 * Dice refresh callback
 */
void CBK_RefreshDice( Widget w, XtPointer pData, XtPointer call_data )
    {
    UNUSED( w );
    UNUSED( pData );
    UNUSED( call_data );

    DICE_Refresh( );
    }


/**
 * About box handler
 */
void CBK_About( Widget w, XtPointer pData, XtPointer call_data )
    {
    char buf[ 256 ];
    UNUSED( w );
    UNUSED( pData );
    UNUSED( call_data );

    /* Print version information */
    snprintf( buf,
              sizeof( buf ),
              "This is Xfrisk version %s, compiled %s at %s",
              VERSION,
              __DATE__,
              __TIME__ );
    UTIL_DisplayComment( buf );
    }


/**
 * State replication function
 */
void CBK_Replicate( const Msgs *pMess, Int32 iType, Int32 iSrc )
    {
    UNUSED( iSrc );

    /*
     * Do the physical replication( send to the server so it will
     * broadcast it ), but only if this callback is being called
     * for an outgoing message( i.e. a replication ).  Otherwise,
     * a message just came in.
     */

    if ( iType == MESS_OUTGOING )
        {
        ( void ) RISK_SendMessage( CLIENTS_GetCommLinkOfClient( CLNT_GetThisClientID( ) ),
                                   pMess );
        }
    }


/**
 * The Generic Callback Function
 * \bug VERY obscure
 */
void CBK_Callback( const Msgs *pMess )
    {
    /* Based on the message from the Model, perform some sort of action */
    switch ( pMess->msgId )
        {
        case MSG_OBJINTUPDATE:
            switch ( pMess->msgObjIntUpdate.iField )
                {
                case CNT_OWNER:
                    /* Recolor the country */
                    if ( pMess->msgObjIntUpdate.iNewValue >= 0 )
                        {
                        COLOR_ColorCountry( pMess->msgObjIntUpdate.iIndex1,
                                            pMess->msgObjIntUpdate.iNewValue );
                        }
                    break;

                case CNT_NUMARMIES:
                    /* Redraw the number of armies on the country.  Since the
                     * country may be lit up because of the notification, we
                     * take pains not to mess with it.  Check to see if it was
                     * lit up, and if so, leave it that way.
                     */

                    UTIL_PrintArmies( pMess->msgObjIntUpdate.iIndex1,
                                      pMess->msgObjIntUpdate.iNewValue,
                                      CLNT_GetLightCountOfCountry( pMess->msgObjIntUpdate.iIndex1 ) > 0 ?
                                      WhitePixel( hDisplay, 0 ) : BlackPixel( hDisplay, 0 ) );
                    break;

                case PLR_ALLOCATION:
                    /*
                     * Two cases: either a player got killed or during
                     * registration phase, a player did not complete the
                     * registration sequence.
                     *
                     * If a player was killed then remove it from the
                     * MsgDst list, otherwise skip all modifications.
                     */
                    if ( pMess->msgObjIntUpdate.iNewValue == ALLOC_NONE )
                        {
                        Int32   iDeadPlayer = pMess->msgObjIntUpdate.iIndex1;
                        Int32   i;
                        Int32   iIndex;
                        Flag    fDone = FALSE;

                        /* Search for the entry to delete it */
                        for ( i = 1; i != iIndexMD && !fDone; i++ )
                            {
                            if ( piMsgDstPlayerID[ i ] == iDeadPlayer )
                                {
                                fDone = TRUE;
                                }
                            }

                        /*
                         * In case player did not complete registration
                         * sequence, skip all modifications
                         */
                        if ( !fDone )
                            {
                            printf( "Player %u did not complete registration\n",
                                    iDeadPlayer );
                            break;
                            }

                        /* We found the entry.  Delete it and move the others down */
                        iIndex = ( i - 1 );

                        MEM_Free( pstrMsgDstCString[ iIndex ] );
                        piMsgDstPlayerID[ iIndex ] = -1;
                        iIndexMD--;

                        for ( i = iIndex; i < iIndexMD; i++ )
                            {
                            pstrMsgDstCString[ i ] = pstrMsgDstCString[ i + 1 ];
                            piMsgDstPlayerID[ i ] = piMsgDstPlayerID[ i + 1 ];
                            }

                        UTIL_RefreshMsgDest( iIndexMD );
                        }
                    /* If the player was created, add it to the message box */
                    else if ( pMess->msgObjIntUpdate.iNewValue == ALLOC_COMPLETE )
                        {
                        char    buf[ 256 ];

                        /* A new player, add this to the list, update listbox */
                        pstrMsgDstCString[ iIndexMD ] =
                            ( CString ) MEM_Alloc( strlen( 
                                                       RISK_GetNameOfPlayer( pMess->msgObjIntUpdate.iIndex1 ) ) + 1 );
                        piMsgDstPlayerID[ iIndexMD ] = pMess->msgObjIntUpdate.iIndex1;
                        strcpy( pstrMsgDstCString[ iIndexMD ],
                                RISK_GetNameOfPlayer( pMess->msgObjIntUpdate.iIndex1 ) );
                        UTIL_RefreshMsgDest( ++iIndexMD );

                        /* Also create a message that tells of the new player */
                        snprintf( buf, sizeof( buf ), "%s ( %s) has joined.",
                                  RISK_GetNameOfPlayer( pMess->msgObjIntUpdate.iIndex1 ),
                                  RISK_GetNameOfSpecies( 
                                      RISK_GetSpeciesOfPlayer( pMess->msgObjIntUpdate.iIndex1 ) ) );
                        UTIL_DisplayMessage( FROM_SERVER, DST_OTHER, buf );
                        }
                    break;

                default:
                    /* A message that we don't care about */
                    ;
                }
            break;

        case MSG_OBJSTRUPDATE:
            switch ( pMess->msgObjStrUpdate.iField )
                {
                case PLR_COLORSTRING:
                    /* Allocate the color */
                    COLOR_StoreNamedColor( pMess->msgObjStrUpdate.strNewValue,
                                           pMess->msgObjStrUpdate.iIndex1 );
                    break;

                default:
                    /* A message that we don't care about */
                    break;
                }
            break;

        default:
            /* Nothing... */
            break;
        }
    }

/* End of file */
